@@ -18,12 +18,15 @@ import PIL.Image as PILImage
 import quaternion
 from tqdm import tqdm
 # ros
-import geometry_msgs.msg as rosmsg
+import geometry_msgs.msg
 import rosbag
 import rospy.rostime
 from rospy.rostime import Duration
 # kapture
 import kapture
+from kapture.core.flatten import flatten
+from kapture.core.Rigs import Rigs
+from kapture.core.Sensors import Sensors
 import kapture.io.csv as kcsv
 import kapture.io.structure
 from kapture.io.records import get_image_fullpath
@@ -62,7 +65,7 @@ class PositionInfo:
 
     def __init__(self,
                  timestamp: rospy.rostime.Time,
-                 pose6d: rosmsg.Pose):
+                 pose6d: geometry_msgs.msg.Pose):
         self.timestamp = timestamp
         self.pose6d = pose6d
 
@@ -75,40 +78,23 @@ class RosBagImporter:
 
     def __init__(self,
                  rosbag_path: str,
-                 rig_sensors_kapture_path: str,
+                 rigs: Optional[Rigs],
+                 sensors: Sensors,
                  kapture_path: str,
                  force_overwrite_existing: bool = False) -> None:
         """
 
         :param rosbag_path: full path to the rosbag file
-        :param rig_sensors_kapture_path: full path to a minimal kapture directory containing the rig and sensors of the
-         multi-cam used to capture the data (images and odometry)
+        :param rigs: rigs of the sensors
+        :param sensors: sensors definition used for the capture
         :param kapture_path: full path to the top kapture directory to save
         :param force_overwrite_existing: silently overwrite kapture files if already exists
         """
-        self._rig_sensors_kapture_path = path.abspath(rig_sensors_kapture_path)
-        kapture_data = kcsv.kapture_from_dir(rig_sensors_kapture_path)
-        if kapture_data.rigs is None or len(kapture_data.rigs) == 0:
-            raise ValueError(f'Rig definition is empty in {rig_sensors_kapture_path}')
-        self._rigs = kapture_data.rigs
-        # Check we have only one rig_id defined
-        rig_id = None
-        for r_id in self._rigs.keys():
-            if rig_id is None:
-                rig_id = r_id
-            elif r_id != rig_id:
-                raise ValueError(f'Found rig_id {r_id} and {rig_id}:'
-                                 f' there should be only one rig_id defined in {kapture_path}')
-        if rig_id is None:
-            raise ValueError(f'No rig defined in {kapture_data}')
-        self._rig_id = rig_id
-        if kapture_data.cameras is None or len(kapture_data.cameras) == 0:
-            raise ValueError(f'Cameras definition is empty in {rig_sensors_kapture_path}')
-        self.cameras = kapture_data.cameras
-        self._sensors = kapture_data.sensors
         if not path.isfile(rosbag_path):
             raise ValueError(f'Rosbag file {rosbag_path} does not exist')
         self._rosbag_path = rosbag_path
+        self._rigs = rigs
+        self._sensors = sensors
         self._kapture_path = kapture_path
         self.logger = logging.getLogger('rosbag')
         self.logger.info(f'Reading rosbag file {rosbag_path} and exporting as Kapture format into {kapture_path}')
@@ -143,12 +129,12 @@ class RosBagImporter:
         os.utime(image_path, (taken_time, taken_time))
         return path.relpath(image_path, self._images_full_path)
 
-    def _find_pose(self, image_stamp: rospy.rostime.Time) -> Optional[rosmsg.Pose]:
+    def _find_pose(self, image_stamp: rospy.rostime.Time) -> Optional[geometry_msgs.msg.Pose]:
         """
         Find the pose that has the smallest time difference with the image timestamp
 
         :param image_stamp: an image time stamp
-        :return: a rosmsg pose if found, none otherwise
+        :return: a geometric pose if found, none otherwise
         """
         if len(self._last_poses) == 0:
             return None
@@ -159,11 +145,13 @@ class RosBagImporter:
             if delta < smallest_delta:
                 smallest_delta = delta
                 chosen_stamp = stamp
-        return self._last_poses[chosen_stamp].pose
+        return self._last_poses[chosen_stamp]
 
-    def import_multi_camera(self, odometry_topic: str,
+    def import_multi_camera(self, odometry_topic: Optional[str],
                             image_topics: Union[str, List[str]],
                             camera_identifiers: Union[str, List[str]],
+                            save_all_positions: bool = True,
+                            find_image_position: bool = True,
                             percent: int = 100):
         """
         Import the rosbag data. Save the images on disk.
@@ -172,6 +160,8 @@ class RosBagImporter:
         :param odometry_topic: the odometry topic to use to compute the trajectory.
         :param image_topics: image topic(s) to import
         :param camera_identifiers: camera identifier(s) corresponding to the image topic(s)
+        :param save_all_positions: save all positions from the odometry topic in the trajectory
+        :param find_image_position: find the closest position for the image and add it in the trajectory
         :param percent: percentage of images to keep.
         """
 
@@ -187,9 +177,13 @@ class RosBagImporter:
         #
 
         # Check that we have an odometry topic
-        if odometry_topic is None or len(odometry_topic) == 0:
-            self.logger.fatal('Please provide an odometry topic')
-            raise ValueError('Missing odometry topic')
+        if odometry_topic and len(odometry_topic) == 0:
+            odometry_topic = None
+        if odometry_topic is None:
+            self.logger.info('No odometry topic')
+            if find_image_position:
+                self.logger.info('Will not find position for the images without odometry')
+                find_image_position = False
         # Check that we have image topics
         if isinstance(image_topics, str):
             image_topics = [image_topics]
@@ -205,10 +199,10 @@ class RosBagImporter:
                               f' {len(camera_identifiers)} cameras for {nb_image_topics} image topics.')
             raise ValueError('Unequal number of image topics and camera identifiers')
 
-        # Make sure the camera identifiers are defined in the rigs kapture
+        # Make sure the camera identifiers are defined in the list of sensors
         for camera_id in camera_identifiers:
-            if self.cameras.get(camera_id) is None:
-                raise ValueError(f'Camera identifier {camera_id} is not defined in {self._rig_sensors_kapture_path}')
+            if self._sensors.get(camera_id) is None:
+                raise ValueError(f'Camera identifier {camera_id} is not defined in {list(self._sensors.keys())}')
         # Create images directories
         image_directory_path = OrderedDict()  # Should be OrderedDict[str, str]
         image_topic_to_cam_id = dict()
@@ -224,7 +218,7 @@ class RosBagImporter:
             all_topics = list(bag_info[1].keys())
             self.logger.debug(f'Topics found {all_topics}')
             # Check topics
-            if odometry_topic not in all_topics:
+            if odometry_topic and odometry_topic not in all_topics:
                 raise ValueError(f'Missing topic {odometry_topic} in Rosbag')
             for image_topic in image_topics:
                 if image_topic not in all_topics:
@@ -258,19 +252,42 @@ class RosBagImporter:
             else:
                 progress_bar = None
             topics_to_import = image_topics.copy()
-            topics_to_import.append(odometry_topic)
+            if odometry_topic:
+                topics_to_import.append(odometry_topic)
             for topic, msg, t in bag.read_messages(topics=topics_to_import):
                 # Consider the data timestamp, not the recording timestamp t
                 stamp = msg.header.stamp
-                if odometry_topic == topic:
-                    self.logger.debug(f"{num_msgs:5d} ODOM STAMP='{stamp.secs}.{stamp.nsecs}'"
-                                      f" msg.poseX='{msg.pose.pose.position.x}'")
-                    self._last_poses[stamp] = msg.pose
+                if topic == odometry_topic:
+                    self.logger.debug(f"{num_msgs:5d} STAMP='{stamp.secs}.{stamp.nsecs}'"
+                                      f" ODOM msg.poseX='{msg.pose.pose.position.x}'")
+                    self._last_poses[stamp] = msg.pose.pose
+                    if save_all_positions:
+                        self.poses_info.append(PositionInfo(stamp, msg.pose.pose))
                 elif topic in image_topics:
-                    self.logger.debug(f"{num_msgs:5d} {topic} STAMP='{stamp.secs}.{stamp.nsecs}'"
-                                      f" HxW={msg.height}x{msg.width}")
+                    self.logger.debug(f"{num_msgs:5d} STAMP='{stamp.secs}.{stamp.nsecs}'"
+                                      f" {topic} HxW={msg.height}x{msg.width}")
                     images_stamp[topic] = stamp
-                    img_buf = np.frombuffer(msg.data, dtype=np.uint8).reshape(msg.height, msg.width)
+                    # TODO: Better image decoding:
+                    #  cv_bridge (depend on OpenCV)
+                    #  ros_numpy: no dependency on OpenCV but not natively installed on melodic
+                    # The below code is adapted from its Image.py file and will work for types like:
+                    # rgb8, rgba8, rgb16, rgba16, bgr8, bgra8, bgr16, bgra16, mono8, mono16
+                    # and the bayer formats but not the OpenCV CvMat types.
+                    dtype = np.uint8
+                    if len(msg.encoding) > 0:
+                        if msg.encoding.endswith("8"):
+                            dtype = np.uint8
+                        elif msg.encoding.endswith("16"):
+                            dtype = np.uint16
+                        else:
+                            raise ValueError(f'Unsupported image encoding {msg.encoding}')
+                    nb_channels = 1
+                    if msg.step > 0:
+                        nb_channels = int(msg.step / msg.width)
+                    if nb_channels > 1:
+                        img_buf = np.frombuffer(msg.data, dtype=dtype).reshape(msg.height, msg.width, nb_channels)
+                    else:
+                        img_buf = np.frombuffer(msg.data, dtype=dtype).reshape(msg.height, msg.width)
                     images_buffer[topic] = img_buf
                 else:
                     self.logger.debug(f"{num_msgs:5d} Ignoring message of topic '{topic}'")
@@ -282,20 +299,27 @@ class RosBagImporter:
                         for topic2, stamp2 in images_stamp.items():
                             delta = (stamp1-stamp2).to_sec()
                             assert (abs(delta) < 0.001), "{topic1} and {topic2} images must have the same timestamp"
-                    # Find the pose
-                    pose6d = self._find_pose(stamp)
-                    if pose6d:
-                        self.poses_info.append(PositionInfo(stamp, pose6d))
+                    pose6d = self._find_pose(stamp) if find_image_position else None
+                    if pose6d or not find_image_position:
                         # Save only some images
                         if ((image_number-1) % count_to_skip) == 0:
                             for image_topic in image_topics:
                                 img = images_buffer[image_topic]
                                 image_directory = image_directory_path[image_topic]
                                 stamp = images_stamp[image_topic]
-                                img_name = path_secure(self._save_image(img, image_directory, image_number, stamp))
-                                dir_last_image_time[image_directory] = stamp.to_sec()
-                                saved_number += 1
-                                self.images_info.append(ImageInfo(img_name, stamp, image_topic_to_cam_id[image_topic]))
+                                try:
+                                    img_name = path_secure(self._save_image(img, image_directory, image_number, stamp))
+                                except SystemError:
+                                    self.logger.exception(f"Could not save image number {image_number}"
+                                                          f" from topic {image_topic}")
+                                else:
+                                    dir_last_image_time[image_directory] = stamp.to_sec()
+                                    saved_number += 1
+                                    self.images_info.append(
+                                        ImageInfo(img_name, stamp, image_topic_to_cam_id[image_topic])
+                                    )
+                            if pose6d:
+                                self.poses_info.append(PositionInfo(stamp, pose6d))
                         else:
                             self.logger.debug(f'    Skipping image number {image_number}')
                         image_number += 1
@@ -309,12 +333,17 @@ class RosBagImporter:
                 os.utime(image_directory, (dir_time, dir_time))
             self.logger.info(f'Saved {saved_number} images')
 
-    def save_to_kapture(self) -> None:
+    def save_to_kapture(self,
+                        trajectory_rig_id: Optional[str] = None) -> None:
         """
         Save the data in kapture format.
+
+        :param trajectory_rig_id: the rig identifier of the trajectory points
         """
         # Convert pose info to trajectories
-        trajectories = kapture.Trajectories()
+        if len(self.poses_info) > 0 and trajectory_rig_id is None:
+            raise ValueError("Must provide rig identifier for trajectory")
+        trajectories = kapture.Trajectories() if len(self.poses_info) > 0 else None
         for pose_info in self.poses_info:
             t = pose_info.timestamp.to_nsec()
             ros_translation = pose_info.pose6d.position
@@ -335,12 +364,14 @@ class RosBagImporter:
             pose6d = kapture.PoseTransform.compose([pose_kapture_from_ros,
                                                     kapture.PoseTransform(rotation, translation).inverse(),
                                                     pose_ros_from_kapture])
-            trajectories[(t, self._rig_id)] = pose6d
+            trajectories[(t, trajectory_rig_id)] = pose6d
+        self.logger.info(f'Saving {len(list(flatten(trajectories)))} poses')
         # Convert image info to kapture image
         records_camera = kapture.RecordsCamera()
         for image_info in self.images_info:
             t = image_info.timestamp.to_nsec()
             records_camera[(t, image_info.camera_name)] = image_info.filename
+        self.logger.info(f'Saving {len(list(flatten(records_camera)))} camera records')
 
         kapture_data = kapture.Kapture(rigs=self._rigs, sensors=self._sensors,
                                        records_camera=records_camera, trajectories=trajectories)
