@@ -7,7 +7,7 @@ All reading and writing operations of kapture objects in CSV like files
 import os
 import os.path as path
 import re
-from typing import Any, List, Optional, Set, Type, Union
+from typing import Any, List, Optional, Set, Type, Union, Dict
 from collections import namedtuple
 import numpy as np
 
@@ -24,6 +24,7 @@ CSV_FILENAMES = {
     kapture.RecordsCamera: path.join('sensors', 'records_camera.txt'),
     kapture.RecordsLidar: path.join('sensors', 'records_lidar.txt'),
     kapture.RecordsWifi: path.join('sensors', 'records_wifi.txt'),
+    kapture.RecordsGnss: path.join('sensors', 'records_gnss.txt'),
     kapture.Points3d: path.join('reconstruction', 'points3d.txt'),
     kapture.Keypoints: path.join('reconstruction', 'keypoints', 'keypoints.txt'),
     kapture.Descriptors: path.join('reconstruction', 'descriptors', 'descriptors.txt'),
@@ -34,6 +35,9 @@ CSV_FILENAMES = {
 
 def get_csv_fullpath(kapture_type: Any, kapture_dirpath: str = '') -> str:
     """
+    Returns the full path to csv kapture file for a given datastructure and root directory.
+    This path is the concatenation of the kapture root path and subpath into kapture into data structure.
+
     :param kapture_type: type of kapture data (kapture.RecordsCamera, kapture.Trajectories, ...)
     :param kapture_dirpath: root kapture path
     :return: full path of csv file for that type of data
@@ -102,7 +106,6 @@ def kapture_format_version(kapture_dirpath: str) -> Optional[str]:
 def float_safe(representation) -> Union[float, None]:
     """
     Safe float cast
-
     https://stackoverflow.com/questions/6330071/safe-casting-in-python
 
     :param representation: to cast
@@ -117,7 +120,6 @@ def float_safe(representation) -> Union[float, None]:
 def float_array_or_none(representation_list) -> Union[List[float], None]:
     """
     Safe cast of list of float representations
-
     https://stackoverflow.com/questions/6330071/safe-casting-in-python
 
     :param representation_list: list of values to convert
@@ -391,13 +393,16 @@ def records_lidar_to_file(filepath: str, records_lidar: kapture.RecordsLidar) ->
         table_to_file(file, table, header=header)
 
 
-def records_lidar_from_file(filepath: str, lidar_ids: Set[str]) -> kapture.RecordsLidar:
+def records_lidar_from_file(
+        filepath: str,
+        lidar_ids: Optional[Set[str]] = None
+) -> kapture.RecordsLidar:
     """
     Reads records_lidar from CSV file.
 
     :param filepath: input file path
-    :param lidar_ids: input set of valid lidar device ids.
-                        If the records_lidar contains unknown devices, they will be ignored.
+    :param lidar_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
+                            If omitted, then it loads all devices.
     :return: Lidar records
     """
     records_lidar = kapture.RecordsLidar()
@@ -405,7 +410,7 @@ def records_lidar_from_file(filepath: str, lidar_ids: Set[str]) -> kapture.Recor
         table = table_from_file(file)
         # timestamp, device_id, point_cloud_path
         for timestamp, device_id, point_cloud_path in table:
-            if device_id not in lidar_ids:
+            if lidar_ids is not None and device_id not in lidar_ids:
                 # just ignore
                 continue
             records_lidar[(int(timestamp), str(device_id))] = point_cloud_path
@@ -432,13 +437,16 @@ def records_wifi_to_file(filepath: str, records_wifi: kapture.RecordsWifi) -> No
         table_to_file(file, table, header=header)
 
 
-def records_wifi_from_file(filepath: str, wifi_ids: Set[str]) -> kapture.RecordsWifi:
+def records_wifi_from_file(
+        filepath: str,
+        wifi_ids: Optional[Set[str]] = None
+) -> kapture.RecordsWifi:
     """
     Reads RecordsWifi from CSV file.
 
     :param filepath: input file path
-    :param wifi_ids: input set of valid wifi device ids.
-                        If the records_wifi contains unknown devices, they will be ignored.
+    :param wifi_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
+                            If omitted, then it loads all devices.
     :return: Wifi records
     """
     records_wifi = kapture.RecordsWifi()
@@ -448,7 +456,7 @@ def records_wifi_from_file(filepath: str, wifi_ids: Set[str]) -> kapture.Records
         cur_timestamp = -1
         # timestamp (ScanEndTime), device_id, BSSID, RSSI, FREQ, SCANTIME, VISIBLENAME
         for timestamp, device_id, bssid, *wifi_params in table:
-            if device_id not in wifi_ids:
+            if wifi_ids is not None and device_id not in wifi_ids:
                 # just ignore
                 continue
             if timestamp != cur_timestamp:
@@ -464,6 +472,54 @@ def records_wifi_from_file(filepath: str, wifi_ids: Set[str]) -> kapture.Records
     return records_wifi
 
 
+########################################################################################################################
+# Records GNSS #########################################################################################################
+def records_gnss_to_file(
+        filepath: str,
+        records_gnss: kapture.RecordsGnss
+) -> None:
+    """
+    Writes records_gnss to file
+
+    :param filepath:
+    :param records_gnss:
+    """
+    assert isinstance(records_gnss, kapture.RecordsGnss)
+    header = '# timestamp, device_id, x, y, z, utc, dop'
+    table = []
+    for timestamp, sensor_id, gnss_record in kapture.flatten(records_gnss, is_sorted=True):
+        table.append([timestamp, sensor_id] + gnss_record.as_list())
+    with open(filepath, 'w') as file:
+        table_to_file(file, table, header=header)
+
+
+def records_gnss_from_file(
+        filepath: str,
+        gnss_ids: Optional[Set[str]] = None
+) -> kapture.RecordsGnss:
+    """
+    Reads RecordsGnss from CSV file.
+
+    :param filepath: input file path
+    :param gnss_ids: input set of valid device ids. Any record of other than the given ones will be ignored.
+                     If omitted, then it loads all devices.
+    :return: GNSS records
+    """
+    records_gnss = kapture.RecordsGnss()
+    with open(filepath) as file:
+        table = table_from_file(file)
+        # timestamp, device_id, x, y, z, utc, dop
+        for timestamp, device_id, x, y, z, utc, dop in table:
+            timestamp = int(timestamp)
+            device_id = str(device_id)
+            if gnss_ids is not None and device_id not in gnss_ids:
+                # just ignore
+                continue
+            records_gnss[timestamp, device_id] = kapture.RecordGnss(x, y, z, utc, dop)
+
+    return records_gnss
+
+
 ########################################################################################################################
 # features #############################################################################################################
 ImageFeatureConfig = namedtuple('ImageFeatureConfig', ['name', 'dtype', 'dsize'])
@@ -573,7 +629,7 @@ def keypoints_to_file(config_filepath: str, keypoints: kapture.Keypoints) -> Non
 
 def keypoints_from_dir(kapture_dirpath: str, images_paths: Optional[Set[str]]) -> kapture.Keypoints:
     """
-    Read and build keypoints from images_filenames if given, or directly from actual files in kapture_dirpath.
+    Reads and builds keypoints from images_filenames if given, or directly from actual files in kapture_dirpath.
 
     :param kapture_dirpath: root path of kapture
     :param images_paths: optional list of image file names
@@ -599,7 +655,7 @@ def descriptors_to_file(config_filepath: str, descriptors: kapture.Descriptors)
 
 def descriptors_from_dir(kapture_dirpath: str, images_paths: Set[str]) -> kapture.Descriptors:
     """
-    Read and build descriptors from images_filenames if given, or directly from actual files in kapture_dirpath.
+    Reads and builds descriptors from images_filenames if given, or directly from actual files in kapture_dirpath.
 
     :param kapture_dirpath: root path of kapture
     :param images_paths: optional list of image file names
@@ -625,7 +681,7 @@ def global_features_to_file(config_filepath: str, global_features: kapture.Globa
 
 def global_features_from_dir(kapture_dirpath: str, images_paths: Set[str]) -> kapture.GlobalFeatures:
     """
-    Read and build Global features from images_filenames if given, or directly from actual files in kapture_dirpath.
+    Reads and builds Global features from images_filenames if given, or directly from actual files in kapture_dirpath.
 
     :param kapture_dirpath: root path of kapture
     :param images_paths: optional list of image file names
@@ -643,7 +699,7 @@ def matches_from_dir(kapture_dirpath: str,
                      image_filenames: Optional[Set[str]] = None,
                      matches_pairsfile_path: Optional[str] = None) -> kapture.Matches:
     """
-    Read and build Matches from images_filenames if given, or directly from actual files in kapture_dirpath.
+    Reads and builds Matches from images_filenames if given, or directly from actual files in kapture_dirpath.
 
     :param kapture_dirpath: root path of kapture
     :param image_filenames: optional list of image file names
@@ -743,7 +799,7 @@ def observations_from_file(
     """
     assert path.basename(observations_filepath) == 'observations.txt'
     assert images_paths_with_keypoints is None or \
-        (isinstance(images_paths_with_keypoints, set) and len(images_paths_with_keypoints) > 0)
+           (isinstance(images_paths_with_keypoints, set) and len(images_paths_with_keypoints) > 0)
     observations = kapture.Observations()
     with open(observations_filepath) as file:
         table = table_from_file(file)
@@ -770,6 +826,7 @@ KAPTURE_ATTRIBUTE_WRITERS = {
     kapture.RecordsCamera: records_camera_to_file,
     kapture.RecordsLidar: records_lidar_to_file,
     kapture.RecordsWifi: records_wifi_to_file,
+    kapture.RecordsGnss: records_gnss_to_file,
     kapture.Keypoints: keypoints_to_file,
     kapture.Descriptors: descriptors_to_file,
     kapture.GlobalFeatures: global_features_to_file,
@@ -784,6 +841,7 @@ KAPTURE_ATTRIBUTE_NAMES = {  # used to list attributes to be saved
     kapture.RecordsCamera: 'records_camera',
     kapture.RecordsLidar: 'records_lidar',
     kapture.RecordsWifi: 'records_wifi',
+    kapture.RecordsGnss: 'records_gnss',
     kapture.Keypoints: 'keypoints',
     kapture.Descriptors: 'descriptors',
     kapture.GlobalFeatures: 'global_features',
@@ -794,7 +852,7 @@ KAPTURE_ATTRIBUTE_NAMES = {  # used to list attributes to be saved
 
 def kapture_to_dir(dirpath: str, kapture_data: kapture.Kapture) -> None:
     """
-    Saves kapture directory.
+    Saves kapture data to given directory.
 
     :param dirpath: input directory root path
     :param kapture_data: input kapture data
@@ -812,26 +870,71 @@ def kapture_to_dir(dirpath: str, kapture_data: kapture.Kapture) -> None:
 
 
 # Kapture Read #########################################################################################################
-def kapture_from_dir(kapture_dirpath: str, matches_pairsfile_path: Optional[str] = None) -> kapture.Kapture:
+# list all data members of kapture.
+KAPTURE_LOADABLE_TYPES = {
+    kapture.Sensors,
+    kapture.Rigs,
+    kapture.Trajectories,
+    kapture.RecordsCamera,
+    kapture.RecordsLidar,
+    kapture.RecordsWifi,
+    kapture.RecordsGnss,
+    kapture.Keypoints,
+    kapture.Descriptors,
+    kapture.GlobalFeatures,
+    kapture.Matches,
+    kapture.Points3d,
+    kapture.Observations,
+}
+
+
+def kapture_from_dir(
+        kapture_dirpath: str,
+        matches_pairsfile_path: Optional[str] = None,
+        skip_list: List[Type[Union[kapture.Rigs,
+                                   kapture.Trajectories,
+                                   kapture.RecordsCamera,
+                                   kapture.RecordsLidar,
+                                   kapture.RecordsWifi,
+                                   kapture.RecordsGnss,
+                                   kapture.Keypoints,
+                                   kapture.Descriptors,
+                                   kapture.GlobalFeatures,
+                                   kapture.Matches,
+                                   kapture.Points3d,
+                                   kapture.Observations,
+
+        ]]] = []
+) -> kapture.Kapture:
     """
-    Reads kapture directory.
+    Reads and return kapture data from directory.
 
     :param kapture_dirpath: kapture directory root path
     :param matches_pairsfile_path: text file in the csv format; where each line is image_name1, image_name2, score
+    :param skip_list: Input option for expert only. Skip the load of specified parts.
     :return: kapture data read
     """
     if not path.isdir(kapture_dirpath):
         raise FileNotFoundError(f'No kapture directory {kapture_dirpath}')
-    csv_filepaths = {dtype: path.join(kapture_dirpath, filename) for dtype, filename in CSV_FILENAMES.items()}
-    data_dirpaths = {dtype: path.join(kapture_dirpath, dir_name) for dtype, dir_name in
-                     kapture.io.features.FEATURES_DATA_DIRNAMES.items()}
+    csv_filepaths = {dtype: path.join(kapture_dirpath, filename)
+                     for dtype, filename in CSV_FILENAMES.items()}
+    data_dirpaths = {dtype: path.join(kapture_dirpath, dir_name)
+                     for dtype, dir_name in kapture.io.features.FEATURES_DATA_DIRNAMES.items()}
+
+    # keep only those in load_only and that exists
+    kapture_data_paths = {**data_dirpaths, **csv_filepaths}  # make sure files take precedence over dirs
+    kapture_loadable_data = {
+        kapture_type
+        for kapture_type in KAPTURE_LOADABLE_TYPES
+        if kapture_type not in skip_list and path.exists(kapture_data_paths[kapture_type])
+    }
 
     kapture_data = kapture.Kapture()
     # sensors
     sensor_ids = None
     sensors_file_path = csv_filepaths[kapture.Sensors]
-    if path.exists(sensors_file_path):
-        logger.debug(f'loading sensors')
+    if sensors_file_path:
+        logger.debug(f'loading sensors {sensors_file_path} ...')
         kapture_data.__version__ = get_version_from_csv_file(sensors_file_path)
         kapture_data.sensors = sensors_from_file(sensors_file_path)
         sensor_ids = set(kapture_data.sensors.keys()) if kapture_data.sensors is not None else set()
@@ -841,88 +944,108 @@ def kapture_from_dir(kapture_dirpath: str, matches_pairsfile_path: Optional[str]
         raise FileNotFoundError(f'File {sensors_file_path} is missing or empty in {kapture_dirpath}')
 
     # rigs
-    if path.exists(csv_filepaths[kapture.Rigs]):
-        logger.debug(f'loading rigs')
+    if kapture.Rigs in kapture_loadable_data:
+        rigs_file_path = csv_filepaths[kapture.Rigs]
+        logger.debug(f'loading rigs {rigs_file_path} ...')
         assert sensor_ids is not None
-        kapture_data.rigs = rigs_from_file(csv_filepaths[kapture.Rigs], sensor_ids)
+        kapture_data.rigs = rigs_from_file(rigs_file_path, sensor_ids)
         # update sensor_ids with rig_id
         sensor_ids.update(kapture_data.rigs.keys())
 
     # trajectories
-    if path.exists(csv_filepaths[kapture.Trajectories]):
-        logger.debug(f'loading trajectories')
+    if kapture.Trajectories in kapture_loadable_data:
+        trajectories_file_path = csv_filepaths[kapture.Trajectories]
+        logger.debug(f'loading trajectories {trajectories_file_path} ...')
         assert sensor_ids is not None
-        kapture_data.trajectories = trajectories_from_file(csv_filepaths[kapture.Trajectories], sensor_ids)
+        kapture_data.trajectories = trajectories_from_file(trajectories_file_path, sensor_ids)
 
     # records camera
-    if path.exists(csv_filepaths[kapture.RecordsCamera]):
-        logger.debug(f'loading images')
+    if kapture.RecordsCamera in kapture_loadable_data:
+        records_camera_file_path = csv_filepaths[kapture.RecordsCamera]
+        logger.debug(f'loading images {records_camera_file_path} ...')
         assert kapture_data.sensors is not None
         camera_sensor_ids = set([sensor_id
                                  for sensor_id in kapture_data.sensors.keys()
                                  if kapture_data.sensors[sensor_id].sensor_type == 'camera'])
         kapture_data.records_camera = records_camera_from_file(csv_filepaths[kapture.RecordsCamera], camera_sensor_ids)
 
-    # records Wifi
-    if path.exists(csv_filepaths[kapture.RecordsWifi]):
-        logger.debug(f'loading wifi')
-        assert kapture_data.sensors is not None
-        wifi_sensor_ids = set([sensor_id
-                               for sensor_id in kapture_data.sensors.keys()
-                               if kapture_data.sensors[sensor_id].sensor_type == 'wifi'])
-        assert wifi_sensor_ids is not None
-        kapture_data.records_wifi = records_wifi_from_file(csv_filepaths[kapture.RecordsWifi], wifi_sensor_ids)
-
     # records lidar
-    if path.exists(csv_filepaths[kapture.RecordsLidar]):
-        logger.debug(f'loading lidar')
+    if kapture.RecordsLidar in kapture_loadable_data:
+        records_lidar_file_path = csv_filepaths[kapture.RecordsLidar]
+        logger.debug(f'loading lidar {records_lidar_file_path} ...')
         assert kapture_data.sensors is not None
         lidar_sensor_ids = set([sensor_id
                                 for sensor_id in kapture_data.sensors.keys()
                                 if kapture_data.sensors[sensor_id].sensor_type == 'lidar'])
         assert lidar_sensor_ids is not None
-        kapture_data.records_lidar = records_lidar_from_file(csv_filepaths[kapture.RecordsLidar], lidar_sensor_ids)
+        kapture_data.records_lidar = records_lidar_from_file(records_lidar_file_path, lidar_sensor_ids)
 
+    # records Wifi
+    if kapture.RecordsWifi in kapture_loadable_data:
+        records_wifi_file_path = csv_filepaths[kapture.RecordsWifi]
+        logger.debug(f'loading wifi {records_wifi_file_path} ...')
+        assert kapture_data.sensors is not None
+        wifi_sensor_ids = set([sensor_id
+                               for sensor_id in kapture_data.sensors.keys()
+                               if kapture_data.sensors[sensor_id].sensor_type == 'wifi'])
+        assert wifi_sensor_ids is not None
+        kapture_data.records_wifi = records_wifi_from_file(records_wifi_file_path, wifi_sensor_ids)
+
+    # records GNSS
+    if kapture.RecordsGnss in kapture_loadable_data:
+        records_gnss_file_path = csv_filepaths[kapture.RecordsGnss]
+        logger.debug(f'loading GNSS {records_gnss_file_path} ...')
+        assert kapture_data.sensors is not None
+        epsg_codes = {sensor_id: sensor.sensor_params[0]
+                      for sensor_id, sensor in kapture_data.sensors.items()
+                      if sensor.sensor_type == 'gnss'}
+        if len(epsg_codes) > 0:
+            kapture_data.records_gnss = records_gnss_from_file(records_gnss_file_path, epsg_codes)
+        else:
+            logger.warning('no declared GNSS sensors: all GNSS data will be ignored')
+
+    # features
     image_filenames = set(image_name
                           for _, _, image_name in
-                          kapture.flatten(kapture_data.records_camera))\
+                          kapture.flatten(kapture_data.records_camera)) \
         if kapture_data.records_camera is not None else set()
 
     # keypoints
-    if path.exists(csv_filepaths[kapture.Keypoints]):
-        logger.debug(f'loading keypoints')
+    if kapture.Keypoints in kapture_loadable_data:
+        logger.debug(f'loading keypoints {data_dirpaths[kapture.Keypoints]} ...')
         assert kapture_data.records_camera is not None
         kapture_data.keypoints = keypoints_from_dir(kapture_dirpath, image_filenames)
 
     # descriptors
-    if path.exists(csv_filepaths[kapture.Descriptors]):
-        logger.debug(f'loading descriptors')
+    if kapture.Descriptors in kapture_loadable_data:
+        logger.debug(f'loading descriptors {data_dirpaths[kapture.Descriptors]} ...')
         assert kapture_data.records_camera is not None
         kapture_data.descriptors = descriptors_from_dir(kapture_dirpath, image_filenames)
 
     # global_features
-    if path.exists(csv_filepaths[kapture.GlobalFeatures]):
-        logger.debug(f'loading global features')
+    if kapture.GlobalFeatures in kapture_loadable_data:
+        logger.debug(f'loading global features {data_dirpaths[kapture.GlobalFeatures]} ...')
         assert kapture_data.records_camera is not None
         kapture_data.global_features = global_features_from_dir(kapture_dirpath, image_filenames)
 
     # matches
-    if path.exists(data_dirpaths[kapture.Matches]):
-        logger.debug(f'loading matches')
+    if kapture.Matches in kapture_loadable_data:
+        logger.debug(f'loading matches {data_dirpaths[kapture.Matches]} ...')
         assert kapture_data.records_camera is not None
         kapture_data.matches = matches_from_dir(kapture_dirpath, image_filenames, matches_pairsfile_path)
 
     # points3d
-    if path.exists(csv_filepaths[kapture.Points3d]):
-        logger.debug(f'loading points 3d')
-        kapture_data.points3d = points3d_from_file(csv_filepaths[kapture.Points3d])
+    if kapture.Points3d in kapture_loadable_data:
+        points3d_file_path = csv_filepaths[kapture.Points3d]
+        logger.debug(f'loading points 3d {points3d_file_path} ...')
+        kapture_data.points3d = points3d_from_file(points3d_file_path)
 
     # observations
-    if path.exists(csv_filepaths[kapture.Observations]):
-        logger.debug(f'loading observations')
+    if kapture.Observations in kapture_loadable_data:
+        observations_file_path = csv_filepaths[kapture.Observations]
+        logger.debug(f'loading observations {observations_file_path} ...')
         assert kapture_data.keypoints is not None
         assert kapture_data.points3d is not None
-        kapture_data.observations = observations_from_file(get_csv_fullpath(kapture.Observations, kapture_dirpath),
-                                                           kapture_data.keypoints)
+        kapture_data.observations = observations_from_file(observations_file_path, kapture_data.keypoints)
 
     return kapture_data
