@@ -25,7 +25,7 @@ logging.basicConfig(format='%(levelname)-8s::%(name)s: %(message)s')
 
 INDEX_FILENAME = 'kapture_dataset_index.yaml'
 DEFAULT_DATASET_PATH = path.normpath(path.abspath('.'))
-DEFAULT_REPOSITORY_URL = 'http://download.europe.naverlabs.com/kapture/'
+DEFAULT_REPOSITORY_URL = 'https://github.com/naver/kapture/raw/master/dataset'
 
 datasets = {}
 
@@ -47,14 +47,16 @@ class Dataset:
                  name: str,
                  url: str,
                  root_path: str,
-                 sub_path: str,
-                 sha256sum: str):
+                 sha256sum: str,
+                 sub_path: Optional[str] = None
+                 ):
         """
         :param name: name of the archive (dataset or part of a dataset)
         :param url: input remote url of the dataset archive (tar).
         :param root_path: input absolute path to root directory where all datasets are installed.
-        :param sub_path: input relative path where to install the dataset.
         :param sha256sum: sha256 sum of the archive file on remote server.
+        :param sub_path: input optional sub path where to install the dataset under root_path/sub_path.
+                         If not given, tar file is extracted directly to root_path
         """
         self._name = name
         # remore url of the archive
@@ -64,7 +66,7 @@ class Dataset:
         # if _success_filepath is present, it means it as already been installed. It contains some debug info
         self._success_filepath = path.join(root_path, name + '.txt')
         # local absolute path where to the install the dataset (multiple dataset might be installed at same path).
-        self._install_local_path = path.join(root_path, sub_path)
+        self._install_local_path = path.join(root_path, sub_path) if sub_path is not None else root_path
         self._sha256sum_archive_remote = sha256sum
 
     def is_archive_valid(self):
